﻿#!/usr/bin/env python
# coding: utf-8

# In[2]:


import stl
from stl.mesh import Mesh
import math
import time
import os
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import ConvexHull, convex_hull_plot_2d


# In[3]:


TIME_START = time.time()
# Unique line highlighting function
def unique_rows(arr): 
    arr = np.ascontiguousarray(arr) 
    unique_arr = np.unique(arr.view([('', arr.dtype)]*arr.shape[1])) 
    return unique_arr.view(arr.dtype).reshape((unique_arr.shape[0], arr.shape[1])) 

# Returns model points sorted by Z
def points_from_stl(stl_model):
    arr = np.array (stl_model.vectors)
    a = np.zeros((len(arr),3))
    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            a[i,:] = arr[i,j,:]
    a = unique_rows(a) 
    a = sorted(a, key=lambda tup: tup[2])
    return np.array(a)

feet = Mesh.from_file('model.stl')

points = points_from_stl(feet)
print(points.shape)
print(points)
   


# In[4]:


# Search for the most distant points
def max_points(points):
    point1 = np.array([])
    point2 = np.array([])
    maxlength = 0

    for i in range(points.shape[0]):
        for j in range(points.shape[0]-i):
            cur_length = math.sqrt((points[i,0]-points[j+i,0])**2+(points[i,1]-points[j+i,1])**2+(points[i,2]-points[j+i,2])**2)
            if (cur_length > maxlength):
                maxlength = cur_length
                point1 = points[i,:]
                point2 = points[j+i,:]
    return point1, point2, maxlength

p1, p2, length = max_points(points)
print(p1)
print(p2)
print(length)


# In[6]:


# Line position between max. remote points
hyp_yz = math.sqrt((p2[2]-p1[2])**2+(p2[1]-p1[1])**2)
hyp_xz = math.sqrt((hyp_yz)**2+(p2[0]-p1[0])**2)
angle_x = -math.atan2((p2[2]-p1[2])/hyp_yz,(p2[1]-p1[1])/hyp_yz)+math.pi/2
angle_y = -math.atan2((hyp_yz)/hyp_xz,(p2[0]-p1[0])/hyp_xz)+math.pi/2
if angle_y < 0:
    angle_y = 2*math.pi+angle_y
if angle_x < 0:
    angle_x = 2*math.pi+angle_x
    
print(math.degrees(angle_x))
print(math.degrees(angle_y))


# In[7]:


# Rotate the model so that the longest line is along the Z axis
feet.rotate([0.5,0,0], 2*math.pi - angle_x, p1)
feet.rotate([0,0.5,0], angle_y, p1)

feet.save('R_model.stl')


# In[81]:


# Upload a model rotated by a long line along the Z axis
#feet_1 = Mesh.from_file('rep_points_rotated1.stl')
feet_1 = feet

# find the points
points_1 = points_from_stl(feet_1)
print(points_1.shape)
print(points_1)


# In[9]:


# Find the most distant points
p1_1, p2_1, length_1 = max_points(points_1)
print(p1_1)
print(p2_1)
print(length_1)

print(p1)
print(p2)


# In[55]:


# Checking the corners
hyp_yz_1 = math.sqrt((p2_1[2]-p1_1[2])**2+(p2_1[1]-p1_1[1])**2)
hyp_xz_1 = math.sqrt((p2_1[2]-p1_1[2])**2+(p2_1[0]-p1_1[0])**2)
angle_x = math.atan2((p2_1[2]-p1_1[2])/hyp_yz_1,(p2_1[1]-p1_1[1])/hyp_yz_1)
angle_y = math.atan2((p2_1[2]-p1_1[2])/hyp_xz_1,(p2_1[0]-p1_1[0])/hyp_xz_1)
print(math.degrees(angle_x))
print(math.degrees(angle_y))


# In[82]:


# cross section search
import matplotlib.pyplot as plt
from scipy.spatial import ConvexHull,convex_hull_plot_2d

coef_low = 0.1
coef_high = 0.4
coef_check = 0.3
len_Z = abs(points_1[0,2]-points_1[-1,2])
wide = 0.03*len_Z

fig = plt.figure()

# section function
def cut(p, l):
	global wide
	Z = np.array([[]])
	for i in range (len(p)):
		if p[i,2] >= l + wide:
			break
		if p[i,2] >= l:
			Z = np.append(Z, p[i,0:2])
	Z = Z.reshape(len(Z)//2,2)
	return Z


# In[83]:


# section below
Z_begin = cut(points_1, coef_check*len_Z + points_1[0,2])
# the sum of the axial distances in the first section
Z_begin_sum = (np.max(Z_begin[:,0]) - np.min(Z_begin[:,0])) + (np.max(Z_begin[:,1]) - np.min(Z_begin[:,1]))

# section above
Z_end = cut(points_1, -coef_check*len_Z + points_1[-1,2])
# the sum of the axial distances in the second section
Z_end_sum = (np.max(Z_end[:,0]) - np.min(Z_end[:,0])) + (np.max(Z_end[:,1]) - np.min(Z_end[:,1]))

# we look where the amount is less, determining the top and bottom
if Z_begin_sum > Z_end_sum:
	Z_low = cut(points_1, -coef_low*len_Z + points_1[-1,2])
	Z_high = cut(points_1, -coef_high*len_Z + points_1[-1,2])
	Z_low_z = -coef_low*len_Z + points_1[-1,2] + wide/2
	Z_high_z = -coef_high*len_Z + points_1[-1,2] + wide/2
else:
	Z_low = cut(points_1, coef_low*len_Z + points_1[0,2])
	Z_high = cut(points_1, coef_high*len_Z + points_1[0,2])
	Z_low_z = coef_low*len_Z + points_1[0,2] + wide/2
	Z_high_z = coef_high*len_Z + points_1[0,2] + wide/2


# In[84]:


# circles - almost circles - almost ellipses
#lower
hull_low = ConvexHull(Z_low)
plt.plot(Z_low[:,0],Z_low[:,1], 'o')
for simplex in hull_low.simplices:
	plt.plot(Z_low[simplex,0], Z_low[simplex,1], 'k-', 'b')

#upper
hull_high = ConvexHull(Z_high)
plt.plot(Z_high[:,0],Z_high[:,1], 'o')
for simplex in hull_high.simplices:
	plt.plot(Z_high[simplex,0], Z_high[simplex,1], 'k-', 'g')

# centers and define them
#lower
p1_2 = [(np.max(Z_low[:,0]) + np.min(Z_low[:,0]))/2, (np.max(Z_low[:,1]) + np.min(Z_low[:,1]))/2, Z_low_z]
plt.plot(p1_2[0], p1_2[1], 'bo')

#upper
p2_2 = [(np.max(Z_high[:,0]) + np.min(Z_high[:,0]))/2,(np.max(Z_high[:,1]) + np.min(Z_high[:,1]))/2, Z_high_z]
plt.plot(p2_2[0], p2_2[1], 'go')

#chart
plt.show()


# In[85]:


p1_2


# In[86]:


# Line position between points
hyp_yz_2 = math.sqrt((p2_2[2]-p1_2[2])**2+(p2_2[1]-p1_2[1])**2)
hyp_xz_2 = math.sqrt((hyp_yz_2)**2+(p2[0]-p1[0])**2)
angle_x_1 = -math.atan2((p2_2[2]-p1_2[2])/hyp_yz_2,(p2_2[1]-p1_2[1])/hyp_yz_2)+math.pi/2
angle_y_1 = -math.atan2((hyp_yz_2)/hyp_xz_2,(p2_2[0]-p1_2[0])/hyp_xz_2)+math.pi/2
if angle_y_1 < 0:
    angle_y_1 = 2*math.pi+angle_y_1
if angle_x_1 < 0:
    angle_x_1 = 2*math.pi+angle_x_1
    
print(math.degrees(angle_x_1))
print(math.degrees(angle_y_1))


# In[87]:


# moving model
feet_1.x -= p1_2[0]
feet_1.y -= p1_2[1]
feet_1.z -= p1_2[2]
#Поворот модели
feet_1.rotate([0.5,0,0], 2*math.pi - angle_x_1, [0,0,0])
feet_1.rotate([0,0.5,0], angle_y_1, [0,0,0])

feet_1.save('R_model.stl')


# In[40]:


# Uploading the resulting model
#feet_2 = Mesh.from_file('rep_points_rotated2.stl')
feet_2 = feet_1

# find the points
points_2 = points_from_stl(feet_2)
print(points_2.shape)
print(points_2)


# In[19]:


# condition for determining the top and bottom of the model
k1 = points_2[0, 2]
k2 = points_2[-1, 2] 

if abs(k1) > abs(k2):
    if k1 < 0:
        feet_2.rotate([0.5,0,0], (math.pi), [0,0,0])
        
elif abs(k1) < abs(k2):
    if k2 < 0:
        feet_2.rotate([0.5,0,0], (math.pi), [0,0,0])
                
feet_2.save('R_final.stl')


# In[20]:


t = (time.time() - TIME_START)/60
print('Script execution time =', t, 'min')

# In[ ]:
